/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Communities System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.core.impl;

import java.util.*;
import java.sql.*;
import org.apache.log4j.*;
import com.silverwrist.util.LocaleFactory;
import com.silverwrist.util.StringUtil;
import com.silverwrist.util.rcache.ReferencedData;
import com.silverwrist.venice.*;
import com.silverwrist.venice.core.*;
import com.silverwrist.venice.db.*;
import com.silverwrist.venice.security.PasswordHash;
import com.silverwrist.venice.security.Capability;
import com.silverwrist.venice.security.DefaultLevels;
import com.silverwrist.venice.security.AuditRecord;

class UserContextImpl implements UserContext, UserBackend
{
  /*--------------------------------------------------------------------------------
   * Static data members
   *--------------------------------------------------------------------------------
   */

  private static Category logger = Category.getInstance(UserContextImpl.class.getName());

  private static final String AUTH_TOKEN_PREFIX = "VQAT:";
  private static final char AUTH_TOKEN_SEP = '|';

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private EngineBackend engine;            // the back end of the engine
  private DataPool datapool;               // the data pool used by this object
  private String remote_addr;              // remote address identifier
  private int uid = -1;                    // the user ID we're using
  private int contactid;                   // ID of our contact information
  private int level;                       // base security level for this user
  private int confirm_num;                 // email confirmation number for this user
  private boolean is_anon = false;         // is this the anonymous user?
  private boolean email_verified = false;  // has email address been verified?
  private String username;                 // the user name we're using
  private java.util.Date created;          // when was this user created? (GMT)
  private java.util.Date last_access;      // when did we last log in? (GMT)
  private String description;              // personal description
  private String my_email = null;          // my email address (cached)
  private String my_pseud = null;          // my pseud (cached)
  private String full_name = null;         // my full name (cached)
  private Locale my_locale = null;         // my default locale (cached)
  private TimeZone my_tz = null;           // my default timezone (cached)
  private Hashtable mru_cache = new Hashtable();  // MRU cache for ReferencedData objects

  /*--------------------------------------------------------------------------------
   * Constructor
   *--------------------------------------------------------------------------------
   */

  UserContextImpl(EngineBackend engine, DataPool datapool)
  {
    this.engine = engine;
    this.datapool = datapool;

  } // end constructor

  /*--------------------------------------------------------------------------------
   * finalize() method
   *--------------------------------------------------------------------------------
   */

  protected void finalize()
  {
    Iterator it = mru_cache.values().iterator();
    while (it.hasNext())
    { // release all our ReferencedData objects
      ReferencedData rd = (ReferencedData)(it.next());
      rd.rd_release();

    } // end while

    engine = null;
    datapool = null;
    username = null;
    created = null;
    last_access = null;
    description = null;
    my_email = null;
    my_pseud = null;
    full_name = null;
    my_locale = null;
    my_tz = null;

  } // end finalize

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private void loadUserData(ResultSet rs) throws SQLException
  {
    uid = rs.getInt("uid");
    username = rs.getString("username");
    // skip field "passhash"
    contactid = rs.getInt("contactid");
    is_anon = rs.getBoolean("is_anon");
    email_verified = rs.getBoolean("verify_email");
    // skip fields "lockout", "access_tries"
    confirm_num = rs.getInt("email_confnum");
    level = rs.getInt("base_lvl");
    created = SQLUtil.getFullDateTime(rs,"created");
    last_access = SQLUtil.getFullDateTime(rs,"lastaccess");
    // skip field "passreminder"
    description = rs.getString("description");

    // purge any "cached" fields that may be left over
    my_email = null;
    my_pseud = null;
    full_name = null;
    my_locale = null;
    my_tz = null;

    if (logger.isDebugEnabled())
    { // spit it all out to debug info
      logger.debug("Loaded: UID " + uid + ", username \"" + username + "\", contactid " + contactid);
      logger.debug("...is_anon " + is_anon + ", email_verified " + email_verified);
      logger.debug("... level " + level);

    } // end if

  } // end loadUserData

  private void loadPrefs(Connection conn) throws SQLException, DataException
  {
    Statement stmt = conn.createStatement();
    ResultSet rs = stmt.executeQuery("SELECT * FROM userprefs WHERE uid = " + uid + ";");

    if (!(rs.next()))
      throw new DataException("cannot find preferences for user");

    if (my_tz==null)
      my_tz = TimeZone.getTimeZone(rs.getString("tzid"));

    if (my_locale==null)
      my_locale = LocaleFactory.createLocale(rs.getString("localeid"));

  } // end loadPrefs

  private void loadPrefs() throws DataException
  {
    Connection conn = null;

    try
    { // call through to lower level function
      conn = datapool.getConnection();
      loadPrefs(conn);

    } // end try
    catch (SQLException e)
    { // translate into DataException, yadda yadda yadda
      logger.error("SQL error reading preferences: " + e.getMessage(),e);
      throw new DataException("Unable to read user preferences: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end loadPrefs

  private void sendEmailConfirmation() throws DataException, EmailException
  {
    if (logger.isDebugEnabled())
      logger.debug("sendEmailConfirmation(): sending to \"" + my_email + "\"");

    // Create the message to be sent.
    String message = engine.getStockMessage("email-confirm");
    if (message==null)
    { // no message defined?  oy!
      logger.error("internal error condition: email-confirm stock message not defined");
      throw new DataException("INTERNAL: email-confirm message not defined");

    } // end if

    message = StringUtil.replaceAllInstances(message,"$USERNAME",username);
    message = StringUtil.replaceAllInstances(message,"$CONFNUM",String.valueOf(confirm_num));

    // Create the emailer and send the message.
    SimpleEmailer emailer = engine.createEmailer();
    emailer.setTo(my_email);
    emailer.setSubject("Venice Email Confirmation");
    emailer.setText(message);
    emailer.send();
    if (logger.isDebugEnabled())
      logger.debug("...email sent!");

  } // end sendEmailConfirmation

  /*--------------------------------------------------------------------------------
   * Implementations from interface UserContext
   *--------------------------------------------------------------------------------
   */

  public int getUID()
  {
    return uid;

  } // end getUID

  public String getUserName()
  {
    return username;

  } // end getUserName

  public int getContactID()
  {
    return contactid;

  } // end getContactID

  public String getDescription()
  {
    return description;

  } // end getDescription

  public boolean isLoggedIn()
  {
    if (logger.isDebugEnabled())
      logger.debug("isLoggedIn(): uid = " + uid + ", is_anon = " + is_anon);
    return ((uid!=-1) && !is_anon);

  } // end is_logged_in

  public boolean isEmailVerified()
  {
    return email_verified;

  } // end isEmailVerified

  public void authenticate(String username, String password) throws AccessError, DataException
  {
    if (isLoggedIn())
    { // already authenticated, can't authenticate again
      logger.error("UserContext already authenticated (with uid " + uid + ")");
      throw new InternalStateError("context already authenticated");

    } // end if

    if (logger.isDebugEnabled())
      logger.debug("authenticate(): authenticating user \"" + username + "\"...");

    Connection conn = null;
    AuditRecord ar = null;
      
    try
    { // look for a user name matching this user record
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery("SELECT * FROM users WHERE username = '"
				       + SQLUtil.encodeString(username) + "';");

      if (!(rs.next()))
      {  // user not found
	logger.error("...user not found");
	ar = new AuditRecord(AuditRecord.LOGIN_FAIL,0,remote_addr,"Bad username: " + username);
	throw new AccessError("The user account you have specified does not exist.  Please try again.");

      } // end if

      int the_uid = rs.getInt("uid");

      if (rs.getBoolean("is_anon"))
      { // can't log in as Anonymous Honyak
	logger.error("...user is the Anonymous Honyak, can't explicitly login");
	ar = new AuditRecord(AuditRecord.LOGIN_FAIL,the_uid,remote_addr,"Anonymous user");
	throw new AccessError("This account cannot be explicitly logged into.  Please try again.");

      } // end if

      if (rs.getBoolean("lockout"))
      { // account locked out
	logger.error("...user is locked out by the Admin");
	ar = new AuditRecord(AuditRecord.LOGIN_FAIL,the_uid,remote_addr,"Account locked out");
	throw new AccessError("This account has been administratively locked out.  Please contact "
			      + "the system administrator for assistance.");

      } // end if

      // hash the password and compare it to the stored password hash
      PasswordHash phash = new PasswordHash(password);
      if (!(phash.equals(rs.getString("passhash"))))
      { // they specified an incorrect password
	// TODO: advance an "invalid" counter and lock the account out if more than X tries
	logger.warn("...invalid password");
	ar = new AuditRecord(AuditRecord.LOGIN_FAIL,the_uid,remote_addr,"Bad password");
	throw new AccessError("The password you have specified is incorrect.  Please try again.");

      } // end if

      if (logger.isDebugEnabled())
	logger.debug("...authenticated");

      // we're authenticated - load the user data into the context
      loadUserData(rs);

      // update the "last access" time in the database
      java.util.Date mydate = new java.util.Date();
      stmt.executeUpdate("UPDATE users SET lastaccess = '" + SQLUtil.encodeDate(mydate)
			 + "' WHERE uid = " + uid + ";");

      // update the "last access" time in this object
      last_access = mydate;

      // an audit record indicating we logged in OK
      ar = new AuditRecord(AuditRecord.LOGIN_OK,the_uid,remote_addr);

      if (logger.isDebugEnabled())
	logger.debug("...context loaded, we're ready :-)");

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error reading user data: " + e.getMessage(),e);
      throw new DataException("unable to access user data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end if

  } // end authenticate

  public void confirmEmail(int conf_num) throws AccessError, DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("confirmEmail(): confirming for UID " + uid);
    if ((email_verified) || Capability.exemptFromEmailVerification(level))
    { // already confirmed
      if (logger.isDebugEnabled())
	logger.debug("...user has either already confirmed or is exempt");
      return;

    } // end if

    AuditRecord ar = null;

    if (conf_num!=confirm_num)
    { // the confirmation number is wrong
      logger.warn("...confirmation number incorrect");
      ar = new AuditRecord(AuditRecord.VERIFY_FAIL,uid,remote_addr,"Invalid confirmation number");
      engine.saveAuditRecord(ar);
      throw new AccessError("Confirmation number is incorrect.  Please try again.");

    } // end if

    Connection conn = null;

    try
    { // get a connection and set the user's status to reflect the verification
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE users SET verify_email = 1, base_lvl = ");
      sql.append(DefaultLevels.afterEmailVerification()).append(" WHERE uid = ").append(uid).append(';');
      stmt.executeUpdate(sql.toString());
      email_verified = true;
      level = DefaultLevels.afterEmailVerification();

      // record an audit message indicating that we verified OK
      ar = new AuditRecord(AuditRecord.VERIFY_OK,uid,remote_addr);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error updating user data: " + e.getMessage(),e);
      throw new DataException("unable to update user data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end if

  } // end confirmEmail

  public void resendEmailConfirmation() throws DataException, EmailException
  {
    if (logger.isDebugEnabled())
      logger.debug("resendEmailConfirmation(): resending for UID " + uid);
    if ((email_verified) || Capability.exemptFromEmailVerification(level))
    { // already confirmed, no need to resend
      if (logger.isDebugEnabled())
	logger.debug("...user has either already confirmed or is exempt");
      return;

    } // end if

    if (my_email==null)
      getContactInfo();  // forces my_email to be updated

    Connection conn = null;
    AuditRecord ar = null;
    try
    { // need to change the user's email confirmation number first
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // generate new confirmation number
      int new_confirm_num = engine.getNewConfirmationNumber();

      // create an SQL statement to reset the user account information, and execute it
      StringBuffer sql = new StringBuffer("UPDATE users SET email_confnum = ");
      sql.append(new_confirm_num).append(" WHERE uid = ").append(uid).append(';');
      stmt.executeUpdate(sql.toString());

      confirm_num = new_confirm_num;  // save changed value

      // now send the email confirmation!
      sendEmailConfirmation();

      // log that a resend was performed...
      ar = new AuditRecord(AuditRecord.RESEND_CONFIRM,uid,remote_addr);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error updating user data: " + e.getMessage(),e);
      throw new DataException("unable to update user data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end if
    
  } // end resendEmailConfirmation

  public ContactInfo getContactInfo() throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getContactInfo() for UID " + uid);

    ContactInfoImpl rc;
    if (contactid>=0)
      rc = new ContactInfoImpl(datapool,contactid);
    else
      rc = new ContactInfoImpl(uid);
    if (my_email==null)
      my_email = rc.getEmail();
    if (full_name==null)
      full_name = rc.getGivenName() + " " + rc.getFamilyName();
    if (my_pseud==null)
      my_pseud = full_name;
    return rc;

  } // end getContactInfo

  public boolean putContactInfo(ContactInfo ci) throws DataException, EmailException
  {
    if (logger.isDebugEnabled())
      logger.debug("putContactInfo() for UID " + uid);

    boolean email_changed = false;
    if ((ci.getOwnerUID()!=uid) || (ci.getOwnerSIGID()>=0))
    { // the contact information is not owned correctly
      logger.error("ContactInfo ownership wrong (it's " + ci.getOwnerUID() + ", " + ci.getOwnerSIGID()
		   + "), should be (" + uid + ", -1)");
      throw new DataException("invalid contact information record");

    } // end if

    if (is_anon)
    { // we can't change Anonymous Honyak this way!
      logger.error("cannot modify anonymous user contact info");
      throw new DataException("unmodifiable contact information");

    } // end if

    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Stashable obj = (Stashable)ci;

      // save the contact information
      obj.stash(conn);

      if (contactid<0)
      { // contact being established for the first time
	contactid = ci.getContactID();
	if (logger.isDebugEnabled())
	  logger.debug("...established initial contact (" + contactid + ") for user");
	my_email = ci.getEmail();
	sendEmailConfirmation();
	email_changed = true;

      } // end if

      full_name = ci.getGivenName() + " " + ci.getFamilyName();  // update this field
      my_pseud = full_name;

      if (my_email==null)  // filling in, this is not necessarily the first time
	my_email = ci.getEmail();
      else if (!(my_email.equals(ci.getEmail())) && !Capability.exemptFromEmailVerification(level))
      { // email address change - need to reconfirm - but choose a new confirmation
	// number and save it in the database first; also turn off the verify flag
	// and set the base level to GLOBAL_UNVERIFIED
	// NOTE: Administrators are exempt from this requirement!

	if (logger.isDebugEnabled())
	  logger.debug("email address changed, need to reconfirm");

	// generate new confirmation number
	int new_confirm_num = engine.getNewConfirmationNumber();

	// create an SQL statement to reset the user account information, and execute it
	StringBuffer sql = new StringBuffer("UPDATE users SET verify_email = 0, email_confnum = ");
	sql.append(new_confirm_num).append(", base_lvl = ").append(DefaultLevels.afterEmailAddressChange());
	sql.append(" WHERE uid = ").append(uid).append(';');
	Statement stmt = conn.createStatement();
	stmt.executeUpdate(sql.toString());

	// save off changed data
	email_verified = false;
	confirm_num = new_confirm_num;
	level = DefaultLevels.afterEmailAddressChange();

	// now send the email confirmation!
	sendEmailConfirmation();

	email_changed = true;

      } // end else if

      ar = new AuditRecord(AuditRecord.USER_CONTACT_INFO,uid,remote_addr,"contactid=" + contactid);

    } // end try
    catch (ClassCastException cce)
    { // we need to be able to coerce the ContactInfo to a Stashable
      logger.error("ContactInfo needs to be a Stashable for this to work");
      throw new DataException("improper contact information record");

    } // end catch
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error updating contact info: " + e.getMessage(),e);
      throw new DataException("unable to access user contact data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end if

    return email_changed;

  } // end putContactInfo

  public UserProfile getProfile(String xusername) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getProfile(\"" + xusername + "\")...");
    Connection conn = null;

    try
    { // retrieve a connection from the data pool
      conn = datapool.getConnection();
      UserProfileImpl prof = new UserProfileImpl(engine,this,conn,xusername,
						 Capability.canSeeHiddenContactFields(level));
      if (logger.isDebugEnabled())
	logger.debug("...found it!");
      return prof;

    } // end catch
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error looking up profile information: " + e.getMessage(),e);
      throw new DataException("Unable to look up user info: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getProfile

  public UserProfile getProfile(int xuid) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getProfile(#" + xuid + ")...");
    Connection conn = null;

    try
    { // retrieve a connection from the data pool
      conn = datapool.getConnection();
      UserProfileImpl prof = new UserProfileImpl(engine,this,conn,xuid,
						 Capability.canSeeHiddenContactFields(level));
      if (logger.isDebugEnabled())
	logger.debug("...found it!");
      return prof;

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error looking up profile information: " + e.getMessage(),e);
      throw new DataException("Unable to look up user info: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getProfile

  public void setPassword(String password, String reminder) throws DataException
  {
    if (is_anon)
    { // trying to change Anonymous Honyak's (nonexistent) password?
      logger.error("cannot change password of anonymous account");
      throw new DataException("The anonymous account cannot change its password.");

    } // end if

    Connection conn = null;
    AuditRecord ar = null;

    try
    { // retrieve a connection from the data pool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      PasswordHash phash = new PasswordHash(password);
      StringBuffer sql = new StringBuffer("UPDATE users SET passhash = '");
      sql.append(phash.toString()).append("', passreminder = '").append(SQLUtil.encodeString(reminder));
      sql.append("' WHERE uid = ").append(uid).append(';');
      stmt.executeUpdate(sql.toString());

      // record an audit record for this user
      ar = new AuditRecord(AuditRecord.PASSWORD_CHANGE,uid,remote_addr);
      
    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error changing password: " + e.getMessage(),e);
      throw new DataException("Unable to set user password: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setPassword

  public void setDescription(String new_descr) throws DataException
  {
    if (is_anon)
    { // trying to change Anonymous Honyak's (nonexistent) description?
      logger.error("cannot change description of anonymous account");
      throw new DataException("The anonymous account cannot change its description.");

    } // end if

    Connection conn = null;

    try
    { // retrieve a connection from the data pool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE users SET description = '");
      sql.append(SQLUtil.encodeString(new_descr)).append("' WHERE uid = ").append(uid).append(';');
      stmt.executeUpdate(sql.toString());

      description = new_descr;  // change stored information
      
    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error changing description: " + e.getMessage(),e);
      throw new DataException("Unable to set user description: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setDescription

  public List getMemberSIGs() throws DataException
  {
    return SIGUserContextImpl.getMemberSIGEntries(engine,this,datapool);

  } // end getMemberSIGs

  public SIGContext getSIGContext(int sigid) throws DataException
  {
    return SIGUserContextImpl.getSIGContext(engine,this,datapool,sigid);

  } // end getSIGContext

  public SIGContext getSIGContext(String alias) throws DataException
  {
    return SIGUserContextImpl.getSIGContext(engine,this,datapool,alias);

  } // end getSIGContext

  public List getRootCategoryList() throws DataException
  {
    return CategoryDescriptorImpl.getTopLevelCategoryList(datapool,Capability.hideHiddenCategories(level));

  } // end getRootCategoryList

  public CategoryDescriptor getCategoryDescriptor(int catid) throws DataException
  {
    return new CategoryDescriptorImpl(datapool,catid,Capability.hideHiddenCategories(level));

  } // end getCategoryDescriptor

  public List searchForSIGs(int field, int mode, String term, int offset, int count) throws DataException
  {
    return SIGUserContextImpl.searchForSIGs(engine,this,datapool,field,mode,term,offset,count);

  } // end searchForSIGs

  public int getSearchSIGCount(int field, int mode, String term) throws DataException
  {
    return SIGUserContextImpl.getSearchSIGCount(this,datapool,field,mode,term);

  } // end getSearchSIGCount

  public List getSIGsInCategory(int catid, int offset, int count) throws DataException
  {
    return SIGUserContextImpl.getSIGsInCategory(engine,this,datapool,catid,offset,count);

  } // end getSIGsInCategory

  public List getSIGsInCategory(CategoryDescriptor cat, int offset, int count) throws DataException
  {
    return SIGUserContextImpl.getSIGsInCategory(engine,this,datapool,cat.getLinkedCategoryID(),offset,count);

  } // end getSIGsInCategory

  public int getNumSIGsInCategory(int catid) throws DataException
  {
    return SIGUserContextImpl.getNumSIGsInCategory(this,datapool,catid);

  } // end getNumSIGsInCategory

  public int getNumSIGsInCategory(CategoryDescriptor cat) throws DataException
  {
    return SIGUserContextImpl.getNumSIGsInCategory(this,datapool,cat.getLinkedCategoryID());

  } // end getNumSIGsInCategory

  public List searchForCategories(int mode, String term, int offset, int count) throws DataException
  {
    return CategoryDescriptorImpl.searchForCategories(datapool,Capability.hideHiddenCategories(level),
						      Capability.showHiddenSearchCategories(level),mode,
						      term,offset,count);

  } // end searchForCategories

  public int getSearchCategoryCount(int mode, String term) throws DataException
  {
    return CategoryDescriptorImpl.getSearchCategoryCount(datapool,Capability.hideHiddenCategories(level),
							 Capability.showHiddenSearchCategories(level),
							 mode,term);

  } // end getSearchCategoryCount

  public SIGContext createSIG(String name, String alias, String language, String synopsis, String rules,
			      String joinkey, int hide_mode) throws DataException, AccessError
  {
    if (!canCreateSIG())
      throw new AccessError("You are not authorized to create new SIGs.");

    // Convert the "hide mode" value into the two hide flags.
    boolean hide_dir = (hide_mode!=SIGContext.HIDE_NONE);
    boolean hide_search = (hide_mode==SIGContext.HIDE_BOTH);

    // Create the new SIG's database entries and internal data.
    SIGData new_sig = SIGCoreData.createSIG(engine,this,datapool,name,alias,uid,language,synopsis,rules,
					    joinkey,hide_dir,hide_search);

    // Create the SIG context we return to the user.
    SIGContext rc = new SIGUserContextImpl(engine,this,datapool,new_sig);

    new_sig.rd_release();  // release the extra reference we have on SIGData

    // And that's it!  You expected lightning bolts maybe? :-)

    return rc;

  } // end createSIG

  public boolean canCreateSIG()
  {
    return (level>=engine.getParamInt(EngineBackend.IP_CREATESIGLVL));

  } // end canCreateSIG

  public List getSideBoxList() throws DataException
  {
    Connection conn = null;
    ArrayList rc = new ArrayList();

    try
    { // retrieve a connection from the data pool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // retrieve the necessary rows from the sideboxes table
      ResultSet rs = stmt.executeQuery("SELECT boxid, sequence, param FROM sideboxes WHERE uid = " + uid
				       + " ORDER BY sequence;");
      while (rs.next())
      { // create the implementation objects and return them all
	SideBoxDescriptor sbd = new SideBoxDescriptorImpl(engine.getMasterSideBoxDescriptor(rs.getInt(1)),
							  rs.getInt(2),rs.getString(3));
	rc.add(sbd);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error getting side box config: " + e.getMessage(),e);
      throw new DataException("unable to retrieve side box config: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return Collections.unmodifiableList(rc);

  } // end getSideBoxList

  public List getConferenceHotlist() throws DataException
  {
    return ConferenceUserContextImpl.getUserHotlist(engine,this,datapool);

  } // end getConferenceHotlist

  public boolean hasAdminAccess()
  {
    return Capability.canAdministerSystem(level);

  } // end hasAdminAccess

  public AdminOperations getAdminInterface() throws AccessError
  {
    if (!(Capability.canAdministerSystem(level)))
    { // you don't have access to get this!
      logger.error("user does not have access to do system admin stuff");
      throw new AccessError("You are not permitted to administer the server.");

    } // end if

    // create the return object
    return new AdminOperationsImpl(engine,this,datapool);

  } // end getAdminInterface

  public Locale getLocale() throws DataException
  {
    if (my_locale==null)
      loadPrefs();

    return my_locale;

  } // end getLocale

  public void setLocale(Locale locale) throws DataException
  {
    Connection conn = null;

    try
    { // retrieve a connection from the data pool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // create the update statement
      StringBuffer sql = new StringBuffer("UPDATE userprefs SET localeid = '");
      sql.append(SQLUtil.encodeString(locale.toString())).append("' WHERE uid = ").append(uid).append(';');

      // execute the statement
      stmt.executeUpdate(sql.toString());

      // replace the locale here
      my_locale = locale;

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error setting user locale: " + e.getMessage(),e);
      throw new DataException("unable to set user locale: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setLocale

  public TimeZone getTimeZone() throws DataException
  {
    if (my_tz==null)
      loadPrefs();

    return my_tz;

  } // end getTimeZone

  public void setTimeZone(TimeZone timezone) throws DataException
  {
    Connection conn = null;

    try
    { // retrieve a connection from the data pool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // create the update statement
      StringBuffer sql = new StringBuffer("UPDATE userprefs SET tzid = '");
      sql.append(SQLUtil.encodeString(timezone.getID())).append("' WHERE uid = ").append(uid).append(';');

      // execute the statement
      stmt.executeUpdate(sql.toString());

      // replace the locale here
      my_tz = timezone;

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error setting user timezone: " + e.getMessage(),e);
      throw new DataException("unable to set user timezone: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setTimeZone

  public String getAuthenticationToken() throws AccessError, DataException
  {
    if (!isLoggedIn())
    { // can't generate an authentication token if we're not authenticated!
      logger.error("UserContext not authenticated, cannot generate auth token");
      throw new AccessError("You cannot generate an authentication token without logging in.");

    } // end if

    // Generate a random authentication string and poke it into the database for this user.
    String tokenauth = engine.generateRandomAuthString();
    Connection conn = null;

    try
    { // retrieve a connection from the data pool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE users SET tokenauth = '");
      sql.append(tokenauth).append("' WHERE uid = ").append(uid).append(';');
      stmt.executeUpdate(sql.toString());

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error setting token authentication string: " + e.getMessage(),e);
      throw new DataException("Unable to set authentication token: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    // Build the full authentication token string value.
    int checkvalue = uid ^ tokenauth.hashCode();
    StringBuffer buf = new StringBuffer(AUTH_TOKEN_PREFIX);
    buf.append(uid).append(AUTH_TOKEN_SEP).append(tokenauth).append(AUTH_TOKEN_SEP).append(checkvalue);
    buf.append(AUTH_TOKEN_SEP);
    return buf.toString();

  } // end getAuthenticationToken

  public boolean authenticateWithToken(String token) throws DataException
  {
    if (isLoggedIn())
    { // already authenticated, can't authenticate again
      logger.error("UserContext already authenticated (with uid " + uid + ")");
      throw new InternalStateError("context already authenticated");

    } // end if

    if (logger.isDebugEnabled())
      logger.debug("decoding authtoken: " + token);

    // Pick apart the authentication token value.
    if (!(token.startsWith(AUTH_TOKEN_PREFIX)))
    { // token parse error
      logger.error("Token parse error: prefix not valid");
      return false;

    } // end if

    int xstart = AUTH_TOKEN_PREFIX.length();
    int xend = token.indexOf(AUTH_TOKEN_SEP,xstart);
    if (xend<0)
    { // could not find the UID separator
      logger.error("Token parse error: UID sep not found");
      return false;

    } // end if

    int pending_uid;
    try
    { // get the user ID
      pending_uid = Integer.parseInt(token.substring(xstart,xend));

    } // end try
    catch (NumberFormatException nfe)
    { // we couldn't parse the UID
      logger.error("Token parse error: invalid UID value");
      return false;

    } // end catch

    xstart = xend + 1;
    xend = token.indexOf(AUTH_TOKEN_SEP,xstart);
    if (xend<0)
    { // could not find the auth string separator
      logger.error("Token parse error: auth string sep not found");
      return false;

    } // end if

    String pending_auth = token.substring(xstart,xend);
    if (!(engine.isValidRandomAuthString(pending_auth)))
    { // the auth string is not valid by the rules under which it was generated
      logger.error("Token parse error: invalid auth string value");
      return false;

    } // end if

    xstart = xend + 1;
    xend = token.indexOf(AUTH_TOKEN_SEP,xstart);
    if (xend<0)
    { // could not find the checkvalue separator
      logger.error("Token parse error: checkvalue sep not found");
      return false;

    } // end if

    int checkvalue;
    try
    { // get the check value
      checkvalue = Integer.parseInt(token.substring(xstart,xend));

    } // end try
    catch (NumberFormatException nfe)
    { // we couldn't parse the checkvalue
      logger.error("Token parse error: invalid checkvalue");
      return false;

    } // end catch

    if (checkvalue!=(pending_uid ^ pending_auth.hashCode()))
    { // the checkvalue does not match what it should - possible corrupted token
      logger.error("Token parse error: checkvalue does not match");
      return false;

    } // end if

    // At this point, we now have a UID and authentication string extracted from the token.
    // Proceed to authenticate.
    if (logger.isDebugEnabled())
      logger.debug("Authenticating user ID#" + pending_uid);

    Connection conn = null;
    AuditRecord ar = null;
      
    try
    { // look for a user record matching this user ID
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery("SELECT * FROM users WHERE uid = " + pending_uid + ";");

      if (!(rs.next()))
      {  // user not found
	logger.error("...user not found");
	ar = new AuditRecord(AuditRecord.LOGIN_FAIL,0,remote_addr,"Bad token UID: " + pending_uid);
	return false;

      } // end if

      if (rs.getBoolean("is_anon"))
      { // can't log in as Anonymous Honyak
	logger.error("...user is the Anonymous Honyak, can't explicitly login");
	ar = new AuditRecord(AuditRecord.LOGIN_FAIL,pending_uid,remote_addr,"Anonymous user");
	return false;

      } // end if

      if (rs.getBoolean("lockout"))
      { // account locked out
	logger.error("...user is locked out by the Admin");
	ar = new AuditRecord(AuditRecord.LOGIN_FAIL,pending_uid,remote_addr,"Account locked out");
	return false;

      } // end if

      // compare the stored token auth value to what we have
      if (!(pending_auth.equals(rs.getString("tokenauth"))))
      { // the auth string is bad - we can't log in
	logger.warn("...invalid authentication string");
	ar = new AuditRecord(AuditRecord.LOGIN_FAIL,pending_uid,remote_addr,"Bad auth-string");
	return false;

      } // end if

      if (logger.isDebugEnabled())
	logger.debug("...authenticated");

      // we're authenticated - load the user data into the context
      loadUserData(rs);

      // update the "last access" time in the database
      java.util.Date mydate = new java.util.Date();
      stmt.executeUpdate("UPDATE users SET lastaccess = '" + SQLUtil.encodeDate(mydate)
			 + "' WHERE uid = " + uid + ";");

      // update the "last access" time in this object
      last_access = mydate;

      // an audit record indicating we logged in OK
      ar = new AuditRecord(AuditRecord.LOGIN_OK,uid,remote_addr);

      if (logger.isDebugEnabled())
	logger.debug("...context loaded, we're ready :-)");

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error reading user data: " + e.getMessage(),e);
      throw new DataException("unable to access user data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end if

    return true;  // token authentication worked!

  } // end authenticateWithToken

  /*--------------------------------------------------------------------------------
   * Implementations from interface UserBackend
   *--------------------------------------------------------------------------------
   */

  public int realUID()
  {
    return uid;

  } // end realUID

  public int realBaseLevel()
  {
    return level;

  } // end realBaseLevel

  public String userRemoteAddress()
  {
    return remote_addr;

  } // end userRemoteAddress

  public String userDefaultPseud() throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("userDefaultPseud() for UID " + uid);
    if (my_pseud==null)
      getContactInfo();
    return my_pseud;

  } // end userDefaultPseud

  public boolean userIsAnonymous()
  {
    return is_anon;

  } // end userIsAnonymous

  public String realUserName()
  {
    return username;

  } // end realUserName

  public String realEmailAddress() throws DataException
  {
    if (my_email==null)
      getContactInfo();
    return my_email;

  } // end realEmailAddress

  public String realFullName() throws DataException
  {
    if (full_name==null)
      getContactInfo();
    return full_name;

  } // end realFullName

  public void saveMRU(String tag, ReferencedData data)
  {
    ReferencedData old = (ReferencedData)(mru_cache.get(tag));
    data.rd_addRef();
    mru_cache.put(tag,data);
    if (old!=null)
      old.rd_release();

  } // end saveMRU

  /*--------------------------------------------------------------------------------
   * Operations private to implementation package
   *--------------------------------------------------------------------------------
   */

  void loadAnonymous(String remote_addr) throws DataException
  {
    Connection conn = null;
    if (logger.isDebugEnabled())
      logger.debug("loadAnonymous() on UserContext: addr " + remote_addr);

    try
    { // retrieve a connection from the data pool
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery("SELECT * FROM users WHERE is_anon = 1;");
      if (!(rs.next()))
      { // could not find the anonymous user - this is an internal error
	logger.error("WTF? anonymous user not found?!?");
	throw new DataException("INTERNAL ERROR - cannot find anonymous user");

      } // end if

      loadUserData(rs);  // load the data for Anonymous Honyak

      this.remote_addr = remote_addr;

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("error loading anonymous user data: " + e.getMessage(),e);
      throw new DataException("unable to access user data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end loadAnonymous

  void loadNewUser(String remote_addr, int uid, int level, String username, int confirm_num,
		   java.util.Date created, java.util.Date last_access)
  {
    if (logger.isDebugEnabled())
      logger.debug("loadNewUser() on UserContext: addr " + remote_addr + ", uid " + uid + ", level "
		   + level + ", username \"" + username + "\"");

    this.remote_addr = remote_addr;
    this.uid = uid;
    this.contactid = -1;
    this.level = level;
    this.username = username;
    this.confirm_num = confirm_num;
    this.created = created;
    this.last_access = last_access;

  } // end loadNewUser
  
} // end class UserContextImpl
